/*
* Copyright (c) 2017-2024 Nils Kohl, Daniel Bauer, Fabian Böhm.
*
* This file is part of HyTeG
* (see https://i10git.cs.fau.de/hyteg/hyteg).
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*
* The entire file was generated with the HyTeG form generator.
*
* Avoid modifying this file. If buggy, consider fixing the generator itself.
*/

#pragma once

#include "core/DataTypes.h"

#include "hyteg/LikwidWrapper.hpp"
#include "hyteg/communication/Syncing.hpp"
#include "hyteg/edgedofspace/EdgeDoFMacroCell.hpp"
#include "hyteg/n1e1functionspace/N1E1MacroCell.hpp"
#include "hyteg/n1e1functionspace/N1E1VectorFunction.hpp"
#include "hyteg/operators/Operator.hpp"
#include "hyteg/p1functionspace/P1Function.hpp"
#include "hyteg/primitivestorage/PrimitiveStorage.hpp"
#include "hyteg/solvers/Smoothables.hpp"
#include "hyteg/sparseassembly/SparseMatrixProxy.hpp"

#define FUNC_PREFIX

namespace hyteg {

namespace operatorgeneration {

/// Linear combination of double curl and mass.
///
/// Geometry map: IdentityMap
///
/// Weak formulation
///
///     u: trial function (space: N1E1)
///     v: test function  (space: N1E1)
///     α: coefficient    (space: Lagrange, degree: 1)
///     β: coefficient    (space: Lagrange, degree: 1)
///
///     ∫ α curl u · curl v + ∫ β u · v
///
/// Strong formulation
///
///     α curl curl u + β u

class N1E1ElementwiseCurlCurlPlusMass
: public Operator< n1e1::N1E1VectorFunction< walberla::float64 >, n1e1::N1E1VectorFunction< walberla::float64 > >,
  public OperatorWithInverseDiagonal< n1e1::N1E1VectorFunction< walberla::float64 > >
{
 public:
   N1E1ElementwiseCurlCurlPlusMass( const std::shared_ptr< PrimitiveStorage >& storage,
                                    size_t                                     minLevel,
                                    size_t                                     maxLevel,
                                    const P1Function< walberla::float64 >&     _alpha,
                                    const P1Function< walberla::float64 >&     _beta );

   void apply( const n1e1::N1E1VectorFunction< walberla::float64 >& src,
               const n1e1::N1E1VectorFunction< walberla::float64 >& dst,
               uint_t                                               level,
               DoFType                                              flag,
               UpdateType                                           updateType = Replace ) const;

   void toMatrix( const std::shared_ptr< SparseMatrixProxy >& mat,
                  const n1e1::N1E1VectorFunction< idx_t >&    src,
                  const n1e1::N1E1VectorFunction< idx_t >&    dst,
                  uint_t                                      level,
                  DoFType                                     flag ) const;

   void computeInverseDiagonalOperatorValues();

   std::shared_ptr< n1e1::N1E1VectorFunction< walberla::float64 > > getInverseDiagonalValues() const;

 protected:
 private:
   /// Kernel type: apply
   /// - quadrature rule: Xiao-Gimbutas 2 | points: 4, degree: 2
   /// - operations per element:
   ///   adds    muls    divs    pows    abs    assignments    function_calls    unknown_ops
   /// ------  ------  ------  ------  -----  -------------  ----------------  -------------
   ///    546     912       1       0      0              0                 0              0
   void apply_macro_3D( walberla::float64* RESTRICT _data_alpha,
                        walberla::float64* RESTRICT _data_beta,
                        walberla::float64* RESTRICT _data_dst,
                        walberla::float64* RESTRICT _data_src,
                        walberla::float64           macro_vertex_coord_id_0comp0,
                        walberla::float64           macro_vertex_coord_id_0comp1,
                        walberla::float64           macro_vertex_coord_id_0comp2,
                        walberla::float64           macro_vertex_coord_id_1comp0,
                        walberla::float64           macro_vertex_coord_id_1comp1,
                        walberla::float64           macro_vertex_coord_id_1comp2,
                        walberla::float64           macro_vertex_coord_id_2comp0,
                        walberla::float64           macro_vertex_coord_id_2comp1,
                        walberla::float64           macro_vertex_coord_id_2comp2,
                        walberla::float64           macro_vertex_coord_id_3comp0,
                        walberla::float64           macro_vertex_coord_id_3comp1,
                        walberla::float64           macro_vertex_coord_id_3comp2,
                        int64_t                     micro_edges_per_macro_edge,
                        walberla::float64           micro_edges_per_macro_edge_float ) const;
   /// Kernel type: toMatrix
   /// - quadrature rule: Xiao-Gimbutas 2 | points: 4, degree: 2
   /// - operations per element:
   ///   adds    muls    divs    pows    abs    assignments    function_calls    unknown_ops
   /// ------  ------  ------  ------  -----  -------------  ----------------  -------------
   ///    510     948       1       0      0              0                 0              3
   void toMatrix_macro_3D( walberla::float64* RESTRICT          _data_alpha,
                           walberla::float64* RESTRICT          _data_beta,
                           idx_t* RESTRICT                      _data_dst,
                           idx_t* RESTRICT                      _data_src,
                           const Cell&                          cell,
                           const uint_t                         level,
                           walberla::float64                    macro_vertex_coord_id_0comp0,
                           walberla::float64                    macro_vertex_coord_id_0comp1,
                           walberla::float64                    macro_vertex_coord_id_0comp2,
                           walberla::float64                    macro_vertex_coord_id_1comp0,
                           walberla::float64                    macro_vertex_coord_id_1comp1,
                           walberla::float64                    macro_vertex_coord_id_1comp2,
                           walberla::float64                    macro_vertex_coord_id_2comp0,
                           walberla::float64                    macro_vertex_coord_id_2comp1,
                           walberla::float64                    macro_vertex_coord_id_2comp2,
                           walberla::float64                    macro_vertex_coord_id_3comp0,
                           walberla::float64                    macro_vertex_coord_id_3comp1,
                           walberla::float64                    macro_vertex_coord_id_3comp2,
                           std::shared_ptr< SparseMatrixProxy > mat,
                           int64_t                              micro_edges_per_macro_edge,
                           walberla::float64                    micro_edges_per_macro_edge_float ) const;
   /// Kernel type: computeInverseDiagonalOperatorValues
   /// - quadrature rule: Xiao-Gimbutas 2 | points: 4, degree: 2
   /// - operations per element:
   ///   adds    muls    divs    pows    abs    assignments    function_calls    unknown_ops
   /// ------  ------  ------  ------  -----  -------------  ----------------  -------------
   ///    261     390       1       0      0              0                 0              0
   void computeInverseDiagonalOperatorValues_macro_3D( walberla::float64* RESTRICT _data_alpha,
                                                       walberla::float64* RESTRICT _data_beta,
                                                       walberla::float64* RESTRICT _data_invDiag_,
                                                       walberla::float64           macro_vertex_coord_id_0comp0,
                                                       walberla::float64           macro_vertex_coord_id_0comp1,
                                                       walberla::float64           macro_vertex_coord_id_0comp2,
                                                       walberla::float64           macro_vertex_coord_id_1comp0,
                                                       walberla::float64           macro_vertex_coord_id_1comp1,
                                                       walberla::float64           macro_vertex_coord_id_1comp2,
                                                       walberla::float64           macro_vertex_coord_id_2comp0,
                                                       walberla::float64           macro_vertex_coord_id_2comp1,
                                                       walberla::float64           macro_vertex_coord_id_2comp2,
                                                       walberla::float64           macro_vertex_coord_id_3comp0,
                                                       walberla::float64           macro_vertex_coord_id_3comp1,
                                                       walberla::float64           macro_vertex_coord_id_3comp2,
                                                       int64_t                     micro_edges_per_macro_edge,
                                                       walberla::float64           micro_edges_per_macro_edge_float ) const;

   std::shared_ptr< n1e1::N1E1VectorFunction< walberla::float64 > > invDiag_;
   P1Function< walberla::float64 >                                  alpha;
   P1Function< walberla::float64 >                                  beta;
};

} // namespace operatorgeneration

} // namespace hyteg
